#!/usr/bin/python3

# Script designed to add some parameters to domain xml to improve
# interaction between vm and vhs. It should not be used when
# another high-level mgmt tool present, such as nova
#
# In particular:
#
# The script adds features that are available for a specific
# operating system, using the libosinfo library in domain.xml,
# see the features section of this script.
#
# The script adds the <vz:prl_nettool apply="yes"/> field, which
# checked by the libvirt to call the prl_nettool utility to set the
# adapter's network parameters inside the guest OS.
#
# The script adds field <vz:oneshot done="yes"/>, given field
# avoids multiple parameter setting.
#

import os
import sys
from lxml import etree

if len(sys.argv) < 3:
    print('Invalid number of arguments', file=sys.stderr)
    sys.exit(1)

INTEL_NESTED_PATH = "/sys/module/kvm_intel/parameters/nested"
AMD_NESTED_PATH = "/sys/module/kvm_amd/parameters/nested"
TARGET_PCIE_ROOT_PORT_COUNT = 16

VZ_NAMESPACE = 'http://www.virtuozzo.com/vhs'
VZ = "{%s}" % VZ_NAMESPACE
vz_ns = {'vz': VZ_NAMESPACE}

def check_nested_virtualization():
    nested_path = None
    vendor = None
    if os.path.exists(INTEL_NESTED_PATH):
        nested_path = INTEL_NESTED_PATH
        vendor = "Intel"
    elif os.path.exists(AMD_NESTED_PATH):
        nested_path = AMD_NESTED_PATH
        vendor = "AMD"

    if not nested_path:
        return [False, vendor]

    with open(nested_path, 'r') as f:
        return [f.read().strip() == '1', vendor]

def extract_vz(root):
    metadata = root.xpath('/domain/metadata')
    if not metadata:
        metadata = etree.SubElement(root.xpath('/domain')[0], "metadata")
    else:
        metadata = metadata[0]
    vz = metadata.xpath('//vz:vz', namespaces=vz_ns)
    if not vz:
        vz = etree.SubElement(metadata, VZ + "vz", nsmap=vz_ns)
        return vz
    return vz[0]

def extract_vz_oneshot(root):
    vz = extract_vz(root)
    oneshot = vz.xpath('//vz:oneshot', namespaces=vz_ns)
    if not oneshot:
        oneshot = etree.SubElement(vz, VZ + "oneshot", nsmap=vz_ns)
        return oneshot
    return oneshot[0]

def add_prl_netool_option(root):
    vz = extract_vz(root);
    prl_netool_option = vz.xpath('//vz:prl_nettool', namespaces=vz_ns)
    if not prl_netool_option:
        prl_netool_option = etree.SubElement(vz, VZ + "prl_nettool", apply="yes")

def add_pcie_root_ports(root):
    devices = root.xpath('/domain/devices')[0]

    port_count = 0
    last_port = None
    pcie_root = None
    for element in devices.iter("controller"):
        if element.get("model") == "pcie-root-port":
            port_count += 1
            last_port = element
        if element.get("model") == "pcie-root":
            pcie_root = element

    if pcie_root is None:
        pcie_root = etree.SubElement(devices, "controller", type="pci", model="pcie-root")

    for i in range(port_count, TARGET_PCIE_ROOT_PORT_COUNT):
        root_port = etree.Element("controller")
        root_port.set("type", "pci")
        root_port.set("model", "pcie-root-port")
        if last_port is not None:
            last_port.addnext(root_port)
            continue
        pcie_root.addnext(root_port)

def oneshot_preprocess(root, is_q35):
    vz = extract_vz_oneshot(root)
    if vz.get('done') == 'yes':
        return

    add_prl_netool_option(root)

    if is_q35:
        add_pcie_root_ports(root)

    vz.set('done', 'yes')

op = sys.argv[2]
if op == 'define' or op == 'create':
    parser = etree.XMLParser(remove_blank_text=True)
    doc = etree.parse(sys.stdin, parser)

    ostype = doc.xpath('/domain/os/type')
    if not ostype:
        print('Element /domain/os/type is mandatory', file=sys.stderr)
        sys.exit(1)
    ostype = ostype[0]

    if not ostype.xpath('boolean(@machine)'):
        # default to q35
        ostype.attrib['machine'] = 'q35'
    is_q35 = 'q35' in ostype.attrib['machine']

    is_nested, vendor = check_nested_virtualization()
    ns = {'libosinfo': 'http://libosinfo.org/xmlns/libvirt/domain/1.0'}
    osidpath = '/domain/metadata/libosinfo:libosinfo/libosinfo:os/@id'
    osid = doc.xpath(osidpath, namespaces=ns)

    is_win = osid and 'microsoft.com/win' in osid[0]

    if is_win:
        domain = doc.xpath('/domain')[0]
        features = domain.xpath('./features')

        if features:
            features = features[0]
        else:
            features = etree.SubElement(domain, 'features')

        hyperv = features.xpath('./hyperv')

        if hyperv:
            hyperv = hyperv[0]
        else:
            hyperv = etree.SubElement(features, 'hyperv')

        if is_nested:
            if vendor != "AMD" and not hyperv.xpath('./evmcs'):
                etree.SubElement(hyperv, 'evmcs', state='on')

            stimer = hyperv.xpath("./stimer")

            if stimer:
                stimer = stimer[0]
            else:
                stimer = etree.SubElement(hyperv, 'stimer', state='on')

            if stimer.attrib['state'] == 'on' and not stimer.xpath('./direct'):
                etree.SubElement(stimer, 'direct', state='on')

        for f in ["frequencies", "tlbflush", "ipi"]:
            if not hyperv.xpath('./' + f):
                etree.SubElement(hyperv, f, state='on')
    oneshot_preprocess(doc, is_q35=is_q35)

    print(etree.tostring(doc, encoding='unicode', pretty_print=True))
